const Chunk = @import("./kym/Chunk.zig");

const Table = @import("./kym/Table.zig");

const coral = @import("coral");

const file = @import("./file.zig");

const tokens = @import("./kym/tokens.zig");

const tree = @import("./kym/tree.zig");

///
/// Fixed-length integer number type.
///
pub const Fixed = i32;

///
/// Floating point real number type.
///
pub const Float = f64;

///
/// Supertype for all numeric types.
///
pub const Numeric = union (enum) {
	fixed: Fixed,
	float: Float,
	vector2: Vector2,
	vector3: Vector3,
};

///
/// Runtime environment virtual machine state.
///
pub const RuntimeEnv = struct {
	allocator: coral.io.Allocator,
	options: Options,
	interned_symbols: SymbolSet,
	locals: LocalList,
	frames: FrameList,

	const FrameList = coral.list.Stack(struct {
		callable: *RuntimeObj,
		locals_top: usize,
		arg_count: u8,
	});

	const LocalList = coral.list.Stack(?*RuntimeObj);

	///
	/// Optional settings for a [RuntimeEnv].
	///
	pub const Options = struct {
		import_access: file.Access = .null,
		print: ?*const Printer = null,
		print_error: ?*const Printer = null,

		///
		/// Byte string printing function.
		///
		pub const Printer = fn ([]const coral.io.Byte) void;
	};

	const SymbolSet = coral.map.StringTable([:0]coral.io.Byte);

	pub fn add_fixed(self: *RuntimeEnv, rhs_fixed: Fixed) RuntimeError!*RuntimeEnv {
		const addable = try self.expect_object(self.pop());

		defer self.release(addable);

		return switch (addable.internal().payload) {
			.fixed => |lhs_fixed| if (coral.math.checked_add(lhs_fixed, rhs_fixed)) |result|
					self.new_fixed(result)
				else
					self.new_float(@as(Float, @floatFromInt(lhs_fixed)) + @as(Float, @floatFromInt(rhs_fixed))),

			.float => |lhs_float| self.new_float(lhs_float + @as(Float, @floatFromInt(rhs_fixed))),

			.vector2 => |lhs_vector2| new: {
				const rhs_scalar = @as(f32, @floatFromInt(rhs_fixed));

				break: new self.new_vector2(lhs_vector2[0] + rhs_scalar, lhs_vector2[1] + rhs_scalar);
			},

			.vector3 => |lhs_vector3| new: {
				const rhs_scalar = @as(f32, @floatFromInt(rhs_fixed));

				break: new self.new_vector3(lhs_vector3[0] + rhs_scalar, lhs_vector3[1] + rhs_scalar, lhs_vector3[2] + rhs_scalar);
			},

			else => self.raise(error.TypeMismatch, "fixed types are not addable with {typename}", .{
				.typename = addable.get_typename(),
			}),
		};
	}

	pub fn add_float(self: *RuntimeEnv, rhs_float: Float) RuntimeError!*RuntimeEnv {
		const addable = try self.expect_object(self.pop());

		defer self.release(addable);

		return switch (addable.internal().payload) {
			.fixed => |lhs_fixed| self.new_float(@as(Float, @floatFromInt(lhs_fixed)) + rhs_float),
			.float => |lhs_float| self.new_float(lhs_float + lhs_float),

			.vector2 => |lhs_vector2| new: {
				const rhs_scalar = @as(f32, @floatCast(rhs_float));

				break: new self.new_vector2(lhs_vector2[0] + rhs_scalar, lhs_vector2[1] + rhs_scalar);
			},

			.vector3 => |lhs_vector3| new: {
				const rhs_scalar = @as(f32, @floatCast(rhs_float));

				break: new self.new_vector3(lhs_vector3[0] + rhs_scalar, lhs_vector3[1] + rhs_scalar, lhs_vector3[2] + rhs_scalar);
			},

			else => self.raise(error.TypeMismatch, "fixed types are not addable with {typename}", .{
				.typename = addable.get_typename(),
			}),
		};
	}

	pub fn add_vector2(self: *RuntimeEnv, rhs_vector2: Vector2) RuntimeError!*RuntimeEnv {
		const addable = try self.expect_object(self.pop());

		defer self.release(addable);

		return if (addable.is_vector2()) |lhs_vector2|
				self.new_vector2(lhs_vector2[0] + rhs_vector2[0], lhs_vector2[1] + rhs_vector2[1])
			else
				self.raise(error.TypeMismatch, "vector2 types are not addable with {typename}", .{
					.typename = addable.get_typename(),
				});
	}

	pub fn add_vector3(self: *RuntimeEnv, rhs_vector3: Vector3) RuntimeError!*RuntimeEnv {
		const addable = try self.expect_object(self.pop());

		defer self.release(addable);

		return if (addable.is_vector3()) |lhs_vector3|
				self.new_vector3(lhs_vector3[0] + rhs_vector3[0], lhs_vector3[1] + rhs_vector3[1], lhs_vector3[2] + rhs_vector3[2])
			else
				self.raise(error.TypeMismatch, "vector3 types are not addable with {typename}", .{
					.typename = addable.get_typename(),
				});
	}

	pub fn arg_get(self: *RuntimeEnv, arg_index: LocalIndex) RuntimeError!*RuntimeEnv {
		const frame = &self.frames.values[self.frames.values.len - 1];
		const args = self.locals.values[frame.locals_top .. (frame.locals_top + frame.arg_count)];

		if (arg_index < args.len) {
			return self.push(args[arg_index]);
		}

		try self.locals.push_one(null);

		return self;
	}

	pub fn call(self: *RuntimeEnv, local_arg_count: LocalIndex) RuntimeError!*RuntimeEnv {
		const callable = try self.expect_object(self.pop());

		defer self.release(callable);

		const result = get_result: {
			try self.frames.push_one(.{
				.locals_top = self.locals.values.len - local_arg_count,
				.callable = callable.acquire(),
				.arg_count = local_arg_count,
			});

			defer {
				const popped_frame = self.frames.pop().?;

				self.release(popped_frame.callable);
				coral.debug.assert(popped_frame.locals_top < self.locals.values.len);

				var to_discard = self.locals.values.len - popped_frame.locals_top;

				while (to_discard != 0) : (to_discard -= 1) {
					self.discard();
				}
			}

			const frame = &self.frames.values[self.frames.values.len - 1];

			break: get_result try switch (frame.callable.internal().payload) {
				.syscall => |syscall| syscall(self),
				.dynamic => |dynamic| dynamic.typeinfo().call(self, dynamic.userdata()),

				else => self.raise(error.TypeMismatch, "{typename} is not callable", .{
					.typename = frame.callable.get_typename(),
				}),
			};
		};

		errdefer {
			if (result) |object| {
				self.release(object);
			}
		}

		try self.locals.push_one(result);

		return self;
	}

	pub fn compare_greater(self: *RuntimeEnv, rhs_comparable: *RuntimeObj) RuntimeError!*RuntimeEnv {
		const lhs_comparable = try self.expect_object(self.pop());

		defer self.release(lhs_comparable);

		return switch (lhs_comparable.internal().payload) {
			.fixed => |lhs_fixed| switch (rhs_comparable.internal().payload) {
				.fixed => |rhs_fixed| self.new_boolean(@as(Float, @floatFromInt(lhs_fixed)) > @as(Float, @floatFromInt(rhs_fixed))),
				.float => |rhs_float| self.new_boolean(@as(Float, @floatFromInt(lhs_fixed)) > rhs_float),

				else => self.raise(error.TypeMismatch, "right-hand {typename} is not comparable", .{
					.typename = rhs_comparable.get_typename(),
				}),
			},

			.float => |lhs_float| switch (rhs_comparable.internal().payload) {
				.float => |rhs_float| self.new_boolean(lhs_float > rhs_float),
				.fixed => |rhs_fixed| self.new_boolean(lhs_float > @as(Float, @floatFromInt(rhs_fixed))),

				else => self.raise(error.TypeMismatch, "right-hand {typename} is not comparable", .{
					.typename = rhs_comparable.get_typename(),
				}),
			},

			else => self.raise(error.TypeMismatch, "left-hand {typename} is not comparable", .{
				.typename = lhs_comparable.get_typename(),
			}),
		};
	}

	pub fn compare_greater_equals(self: *RuntimeEnv, rhs_comparable: *RuntimeObj) RuntimeError!*RuntimeEnv {
		const lhs_comparable = try self.expect_object(self.pop());

		defer self.release(lhs_comparable);

		return switch (lhs_comparable.internal().payload) {
			.fixed => |lhs_fixed| switch (rhs_comparable.internal().payload) {
				.fixed => |rhs_fixed| self.new_boolean(@as(Float, @floatFromInt(lhs_fixed)) >= @as(Float, @floatFromInt(rhs_fixed))),
				.float => |rhs_float| self.new_boolean(@as(Float, @floatFromInt(lhs_fixed)) >= rhs_float),

				else => self.raise(error.TypeMismatch, "right-hand {typename} is not comparable", .{
					.typename = rhs_comparable.get_typename(),
				}),
			},

			.float => |lhs_float| switch (rhs_comparable.internal().payload) {
				.float => |rhs_float| self.new_boolean(lhs_float >= rhs_float),
				.fixed => |rhs_fixed| self.new_boolean(lhs_float >= @as(Float, @floatFromInt(rhs_fixed))),

				else => self.raise(error.TypeMismatch, "right-hand {typename} is not comparable", .{
					.typename = rhs_comparable.get_typename(),
				}),
			},

			else => self.raise(error.TypeMismatch, "left-hand {typename} is not comparable", .{
				.typename = lhs_comparable.get_typename(),
			}),
		};
	}

	pub fn compare_less(self: *RuntimeEnv, rhs_comparable: *RuntimeObj) RuntimeError!*RuntimeEnv {
		const lhs_comparable = try self.expect_object(self.pop());

		defer self.release(lhs_comparable);

		return switch (lhs_comparable.internal().payload) {
			.fixed => |lhs_fixed| switch (rhs_comparable.internal().payload) {
				.fixed => |rhs_fixed| self.new_boolean(@as(Float, @floatFromInt(lhs_fixed)) < @as(Float, @floatFromInt(rhs_fixed))),
				.float => |rhs_float| self.new_boolean(@as(Float, @floatFromInt(lhs_fixed)) < rhs_float),

				else => return self.raise(error.TypeMismatch, "right-hand {typename} is not comparable", .{
					.typename = rhs_comparable.get_typename(),
				}),
			},

			.float => |lhs_float| switch (rhs_comparable.internal().payload) {
				.float => |rhs_float| self.new_boolean(lhs_float < rhs_float),
				.fixed => |rhs_fixed| self.new_boolean(lhs_float < @as(Float, @floatFromInt(rhs_fixed))),

				else => return self.raise(error.TypeMismatch, "right-hand {typename} is not comparable", .{
					.typename = rhs_comparable.get_typename(),
				}),
			},

			else => return self.raise(error.TypeMismatch, "left-hand {typename} is not comparable", .{
				.typename = lhs_comparable.get_typename(),
			}),
		};
	}

	pub fn compare_less_equals(self: *RuntimeEnv, rhs_comparable: *RuntimeObj) RuntimeError!*RuntimeEnv {
		const lhs_comparable = try self.expect_object(self.pop());

		defer self.release(lhs_comparable);

		return switch (lhs_comparable.internal().payload) {
			.fixed => |lhs_fixed| switch (rhs_comparable.internal().payload) {
				.fixed => |rhs_fixed| self.new_boolean(@as(Float, @floatFromInt(lhs_fixed)) <= @as(Float, @floatFromInt(rhs_fixed))),
				.float => |rhs_float| self.new_boolean(@as(Float, @floatFromInt(lhs_fixed)) <= rhs_float),

				else => return self.raise(error.TypeMismatch, "right-hand {typename} is not comparable", .{
					.typename = rhs_comparable.get_typename(),
				}),
			},

			.float => |lhs_float| switch (rhs_comparable.internal().payload) {
				.float => |rhs_float| self.new_boolean(lhs_float <= rhs_float),
				.fixed => |rhs_fixed| self.new_boolean(lhs_float <= @as(Float, @floatFromInt(rhs_fixed))),

				else => return self.raise(error.TypeMismatch, "right-hand {typename} is not comparable", .{
					.typename = rhs_comparable.get_typename(),
				}),
			},

			else => return self.raise(error.TypeMismatch, "left-hand {typename} is not comparable", .{
				.typename = lhs_comparable.get_typename(),
			}),
		};
	}

	pub fn concat(self: *RuntimeEnv, local_concat_count: LocalIndex) RuntimeError!*RuntimeEnv {
		if (local_concat_count == 0) {
			return self.new_string("");
		}

		const concated_buffer = build_buffer: {
			var concat_buffer = coral.list.ByteStack.init(self.allocator);

			errdefer concat_buffer.deinit();

			const concat_values = self.locals.values[(self.locals.values.len - local_concat_count) .. self.locals.values.len];

			for (concat_values) |concat_value| {
				const string = (try (try self.push(concat_value)).to_string()).pop().?;

				defer self.release(string);

				try concat_buffer.push_all(string.is_string().?);
			}

			{
				var concated_value_count = concat_values.len;

				while (concated_value_count != 0) : (concated_value_count -= 1) {
					self.discard();
				}
			}

			break: build_buffer try concat_buffer.pack();
		};

		errdefer self.allocator.deallocate(concated_buffer);

		const string = try RuntimeObj.allocate(self.allocator, .{
			.ref_count = 1,

			.payload = .{
				.string = .{
					.ptr = concated_buffer.ptr,
					.len = @intCast(concated_buffer.len),
				},
			},
		});

		errdefer self.release(string);

		try self.locals.push_one(string);

		return self;
	}

	///
	/// Deinitializes `self`, freeing all allocated virtual machine resources.
	///
	pub fn deinit(self: *RuntimeEnv) void {
		while (self.locals.pop()) |local| {
			if (local) |ref| {
				self.release(ref);
			}
		}

		{
			var iterable = self.interned_symbols.as_iterable();

			while (iterable.next()) |entry| {
				self.allocator.deallocate(entry.value);
			}
		}

		self.interned_symbols.deinit();
		self.locals.deinit();
		self.frames.deinit();
	}

	pub fn discard(self: *RuntimeEnv) void {
		if (self.pop()) |popped| {
			self.release(popped);
		}
	}

	pub fn div_fixed(self: *RuntimeEnv, rhs_fixed: Fixed) RuntimeError!*RuntimeEnv {
		if (rhs_fixed == 0) {
			return self.raise(error.BadOperation, "cannot divide by zero", .{});
		}

		const addable = try self.expect_object(self.pop());

		defer self.release(addable);

		return switch (addable.internal().payload) {
			.fixed => |lhs_fixed| self.new_float(@as(Float, @floatFromInt(lhs_fixed)) / @as(Float, @floatFromInt(rhs_fixed))),
			.float => |lhs_float| self.new_float(lhs_float / @as(Float, @floatFromInt(rhs_fixed))),

			.vector2 => |lhs_vector2| new: {
				const rhs_scalar = @as(f32, @floatFromInt(rhs_fixed));

				break: new self.new_vector2(lhs_vector2[0] / rhs_scalar, lhs_vector2[1] / rhs_scalar);
			},

			.vector3 => |lhs_vector3| new: {
				const rhs_scalar = @as(f32, @floatFromInt(rhs_fixed));

				break: new self.new_vector3(lhs_vector3[0] / rhs_scalar, lhs_vector3[1] / rhs_scalar, lhs_vector3[2] / rhs_scalar);
			},

			else => self.raise(error.TypeMismatch, "fixed types are not divisible with {typename}", .{
				.typename = addable.get_typename(),
			}),
		};
	}

	pub fn div_float(self: *RuntimeEnv, rhs_float: Float) RuntimeError!*RuntimeEnv {
		if (rhs_float == 0) {
			return self.raise(error.BadOperation, "cannot divide by zero", .{});
		}

		const addable = try self.expect_object(self.pop());

		defer self.release(addable);

		return switch (addable.internal().payload) {
			.fixed => |lhs_fixed| self.new_float(@as(Float, @floatFromInt(lhs_fixed)) / rhs_float),
			.float => |lhs_float| self.new_float(lhs_float / lhs_float),

			.vector2 => |lhs_vector2| new: {
				const rhs_scalar = @as(f32, @floatCast(rhs_float));

				break: new self.new_vector2(lhs_vector2[0] / rhs_scalar, lhs_vector2[1] / rhs_scalar);
			},

			.vector3 => |lhs_vector3| new: {
				const rhs_scalar = @as(f32, @floatCast(rhs_float));

				break: new self.new_vector3(lhs_vector3[0] / rhs_scalar, lhs_vector3[1] / rhs_scalar, lhs_vector3[2] / rhs_scalar);
			},

			else => self.raise(error.TypeMismatch, "fixed types are not divisible with {typename}", .{
				.typename = addable.get_typename(),
			}),
		};
	}

	pub fn div_vector2(self: *RuntimeEnv, rhs_vector2: Vector2) RuntimeError!*RuntimeEnv {
		if (rhs_vector2[0] == 0 or rhs_vector2[1] == 0) {
			return self.raise(error.BadOperation, "cannot divide by zero", .{});
		}

		const addable = try self.expect_object(self.pop());

		defer self.release(addable);

		return if (addable.is_vector2()) |lhs_vector2|
				self.new_vector2(lhs_vector2[0] / rhs_vector2[0], lhs_vector2[1] / rhs_vector2[1])
			else
				self.raise(error.TypeMismatch, "vector2 types are not divisible with {typename}", .{
					.typename = addable.get_typename(),
				});
	}

	pub fn div_vector3(self: *RuntimeEnv, rhs_vector3: Vector3) RuntimeError!*RuntimeEnv {
		if (rhs_vector3[0] == 0 or rhs_vector3[1] == 0 or rhs_vector3[2] == 0) {
			return self.raise(error.BadOperation, "cannot divide by zero", .{});
		}

		const addable = try self.expect_object(self.pop());

		defer self.release(addable);

		return if (addable.is_vector3()) |lhs_vector3|
				self.new_vector3(lhs_vector3[0] / rhs_vector3[0], lhs_vector3[1] / rhs_vector3[1], lhs_vector3[2] / rhs_vector3[2])
			else
				self.raise(error.TypeMismatch, "vector3 types are not divisible with {typename}", .{
					.typename = addable.get_typename(),
				});
	}

	pub fn equals_nil(self: *RuntimeEnv) RuntimeError!*RuntimeEnv {
		if (self.pop()) |lhs_object| {
			defer self.release(lhs_object);

			return self.new_boolean(false);
		}

		return self.new_boolean(true);
	}

	pub fn equals_object(self: *RuntimeEnv, rhs_object: *RuntimeObj) RuntimeError!*RuntimeEnv {
		if (self.pop()) |lhs_object| {
			defer self.release(lhs_object);

			return self.new_boolean(lhs_object.equals(rhs_object));
		}

		return self.new_boolean(false);
	}

	pub fn expect_dynamic(self: *RuntimeEnv, value: *RuntimeObj, typeinfo: *const Typeinfo) RuntimeError![]coral.io.Byte {
		return value.is_dynamic(typeinfo) orelse self.raise(error.TypeMismatch, "expected dynamic object, not {typename}", .{
			.typename = value.get_typename(),
		});
	}

	pub fn expect_float(self: *RuntimeEnv, value: *RuntimeObj) RuntimeError!Float {
		return switch (value.internal().payload) {
			.fixed => |fixed| @floatFromInt(fixed),
			.float => |float| float,

			else => self.raise(error.TypeMismatch, "expected float type, not {typename}", .{
				.typename = value.get_typename(),
			}),
		};
	}

	pub fn expect_fixed(self: *RuntimeEnv, value: *RuntimeObj) RuntimeError!Fixed {
		return value.is_fixed() orelse self.raise(error.TypeMismatch, "expected fixed type, not {typename}", .{
			.typename = value.get_typename(),
		});
	}

	pub fn expect_numeric(self: *RuntimeEnv, value: *RuntimeObj) RuntimeError!Numeric {
		return value.is_numeric() orelse self.raise(error.TypeMismatch, "expected numeric type, not {typename}", .{
			.typename = value.get_typename(),
		});
	}

	pub fn expect_object(self: *RuntimeEnv, value: ?*RuntimeObj) RuntimeError!*RuntimeObj {
		return value orelse self.raise(error.TypeMismatch, "expected object type, not nil", .{});
	}

	pub fn expect_string(self: *RuntimeEnv, value: *RuntimeObj) RuntimeError![]const coral.io.Byte {
		return value.is_string() orelse self.raise(error.TypeMismatch, "expected string type, not {typename}", .{
			.typename = value.get_typename(),
		});
	}

	pub fn expect_symbol(self: *RuntimeEnv, value: *RuntimeObj) RuntimeError![*:0]const coral.io.Byte {
		return value.is_symbol() orelse self.raise(error.TypeMismatch, "expected symbol type, not {typename}", .{
			.typename = value.get_typename(),
		});
	}

	pub fn index_get(self: *RuntimeEnv, index: *RuntimeObj) RuntimeError!*RuntimeEnv {
		const vector_max = 3;
		const indexable = try self.expect_object(self.pop());

		defer self.release(indexable);

		switch (indexable.internal().payload) {
			.vector2 => |vector2| {
				const swizzle_symbol = try self.expect_symbol(index);
				var swizzle_buffer = [_]f32{0} ** vector_max;
				var swizzle_count = @as(usize, 0);

				while (true) : (swizzle_count += 1) {
					if (swizzle_count > swizzle_buffer.len) {
						try self.locals.push_one(null);

						return self;
					}

					swizzle_buffer[swizzle_count] = switch (swizzle_symbol[swizzle_count]) {
						'x' => vector2[0],
						'y' => vector2[1],

						0 => return switch (swizzle_count) {
							1 => self.new_float(swizzle_buffer[0]),
							2 => self.new_vector2(swizzle_buffer[0], swizzle_buffer[1]),
							3 => self.new_vector3(swizzle_buffer[0], swizzle_buffer[1], swizzle_buffer[2]),
							else => unreachable,
						},

						else => return self.raise(error.BadOperation, "no such vector2 index or swizzle: `{index}`", .{
							.index = coral.io.slice_sentineled(@as(coral.io.Byte, 0), swizzle_symbol),
						}),
					};
				}
			},

			.vector3 => |vector3| {
				const swizzle_symbol = try self.expect_symbol(index);
				var swizzle_buffer = [_]f32{0} ** vector_max;
				var swizzle_count = @as(usize, 0);

				while (true) : (swizzle_count += 1) {
					if (swizzle_count > swizzle_buffer.len) {
						try self.locals.push_one(null);

						return self;
					}

					swizzle_buffer[swizzle_count] = switch (swizzle_symbol[swizzle_count]) {
						'x' => vector3[0],
						'y' => vector3[1],
						'z' => vector3[2],

						0 => return switch (swizzle_count) {
							1 => self.new_float(swizzle_buffer[0]),
							2 => self.new_vector2(swizzle_buffer[0], swizzle_buffer[1]),
							3 => self.new_vector3(swizzle_buffer[0], swizzle_buffer[1], swizzle_buffer[2]),
							else => unreachable,
						},

						else => return self.raise(error.BadOperation, "no such vector3 index or swizzle: `{index}`", .{
							.index = coral.io.slice_sentineled(@as(coral.io.Byte, 0), swizzle_symbol),
						}),
					};
				}
			},

			.dynamic => |dynamic| {
				const value = try dynamic.typeinfo().get(self, dynamic.userdata(), index);

				errdefer {
					if (value) |object| {
						self.release(object);
					}
				}

				try self.locals.push_one(value);

				return self;
			},

			else => return self.raise(error.TypeMismatch, "{typename} is not get-indexable", .{
				.typename = indexable.get_typename(),
			}),
		}
	}

	pub fn index_set(self: *RuntimeEnv, index: *RuntimeObj, value: ?*RuntimeObj) RuntimeError!void {
		const indexable = try self.expect_object(self.pop());

		defer self.release(indexable);

		try switch (indexable.internal().payload) {
			.dynamic => |dynamic| dynamic.typeinfo().set(self, dynamic.userdata(), index, value),

			else => self.raise(error.TypeMismatch, "{typename} is not set-indexable", .{
				.typename = indexable.get_typename(),
			}),
		};
	}

	pub fn import(self: *RuntimeEnv, file_path: file.Path) RuntimeError!*RuntimeEnv {
		{
			var callable = new_chunk: {
				const file_name = file_path.get_string();

				const file_data =
					(try file.allocate_and_load(self.allocator, self.options.import_access, file_path)) orelse {
						return self.raise(error.BadOperation, "failed to open or read `{name}`", .{
							.name = file_name,
						});
					};

				defer self.allocator.deallocate(file_data);

				var root = try tree.Root.init(self.allocator);

				defer root.deinit();

				{
					var stream = tokens.Stream{.source = file_data};

					root.parse(&stream) catch |parse_error| {
						for (root.error_messages.values) |error_message| {
							self.print(error_message);
						}

						return self.raise(parse_error, "failed to parse `{name}`", .{.name = file_name});
					};
				}

				var chunk = try Chunk.init(self, file_name, &root.environment);

				errdefer chunk.deinit(self);

				break: new_chunk (try self.new_dynamic(coral.io.bytes_of(&chunk), Chunk.typeinfo)).pop().?;
			};

			errdefer self.release(callable);

			try self.locals.push_one(callable);
		}

		return self.call(0);
	}

	///
	/// Attempts to initialize a new virtual machine instance with `allocator` as the allocation strategy, `frames_max`
	/// as the maximum call depth, and `options` as the optional values to pass to the VM state.
	///
	/// This function will fail if `allocator` cannot allocate the memory required to create the runtime environment.
	///
	pub fn init(allocator: coral.io.Allocator, frames_max: u32, options: Options) coral.io.AllocationError!RuntimeEnv {
		var frames = FrameList.init(allocator);

		errdefer frames.deinit();

		try frames.grow(frames_max);

		var locals = LocalList.init(allocator);

		errdefer locals.deinit();

		try locals.grow(local_max * frames_max);

		return .{
			.interned_symbols = SymbolSet.init(allocator, .{}),
			.locals = locals,
			.frames = frames,
			.allocator = allocator,
			.options = options,
		};
	}

	const LocalIndex = u8;

	const local_max = coral.math.max_int(@typeInfo(LocalIndex).Int);

	pub fn local_get(self: *RuntimeEnv, index: LocalIndex) RuntimeError!*RuntimeEnv {
		const frame = &self.frames.values[self.frames.values.len - 1];

		return self.push(self.locals.values[frame.locals_top + index]);
	}

	pub fn local_set(self: *RuntimeEnv, index: LocalIndex, value: ?*RuntimeObj) void {
		const frame = &self.frames.values[self.frames.values.len - 1];
		const local = &self.locals.values[frame.locals_top + index];

		if (local.*) |previous_local| {
			self.release(previous_local);
		}

		local.* = if (value) |object| object.acquire() else null;
	}

	pub fn local_count(self: *RuntimeEnv) LocalIndex {
		const frame = &self.frames.values[self.frames.values.len - 1];

		return @intCast(self.locals.values[frame.locals_top ..].len);
	}

	pub fn mul_fixed(self: *RuntimeEnv, rhs_fixed: Fixed) RuntimeError!*RuntimeEnv {
		const addable = try self.expect_object(self.pop());

		defer self.release(addable);

		return switch (addable.internal().payload) {
			.fixed => |lhs_fixed| if (coral.math.checked_mul(lhs_fixed, rhs_fixed)) |result|
					self.new_fixed(result)
				else
					self.new_float(@as(Float, @floatFromInt(lhs_fixed)) * @as(Float, @floatFromInt(rhs_fixed))),

			.float => |lhs_float| self.new_float(lhs_float + @as(Float, @floatFromInt(rhs_fixed))),

			.vector2 => |lhs_vector2| new: {
				const rhs_scalar = @as(f32, @floatFromInt(rhs_fixed));

				break: new self.new_vector2(lhs_vector2[0] * rhs_scalar, lhs_vector2[1] * rhs_scalar);
			},

			.vector3 => |lhs_vector3| new: {
				const rhs_scalar = @as(f32, @floatFromInt(rhs_fixed));

				break: new self.new_vector3(lhs_vector3[0] * rhs_scalar, lhs_vector3[1] * rhs_scalar, lhs_vector3[2] * rhs_scalar);
			},

			else => self.raise(error.TypeMismatch, "fixed types are not multiplicable with {typename}", .{
				.typename = addable.get_typename(),
			}),
		};
	}

	pub fn mul_float(self: *RuntimeEnv, rhs_float: Float) RuntimeError!*RuntimeEnv {
		const addable = try self.expect_object(self.pop());

		defer self.release(addable);

		return switch (addable.internal().payload) {
			.fixed => |lhs_fixed| self.new_float(@as(Float, @floatFromInt(lhs_fixed)) * rhs_float),
			.float => |lhs_float| self.new_float(lhs_float * lhs_float),

			.vector2 => |lhs_vector2| new: {
				const rhs_scalar = @as(f32, @floatCast(rhs_float));

				break: new self.new_vector2(lhs_vector2[0] * rhs_scalar, lhs_vector2[1] * rhs_scalar);
			},

			.vector3 => |lhs_vector3| new: {
				const rhs_scalar = @as(f32, @floatCast(rhs_float));

				break: new self.new_vector3(lhs_vector3[0] * rhs_scalar, lhs_vector3[1] + rhs_scalar, lhs_vector3[2] * rhs_scalar);
			},

			else => self.raise(error.TypeMismatch, "fixed types are not multiplicable with {typename}", .{
				.typename = addable.get_typename(),
			}),
		};
	}

	pub fn mul_vector2(self: *RuntimeEnv, rhs_vector2: Vector2) RuntimeError!*RuntimeEnv {
		const addable = try self.expect_object(self.pop());

		defer self.release(addable);

		return if (addable.is_vector2()) |lhs_vector2|
				self.new_vector2(lhs_vector2[0] * rhs_vector2[0], lhs_vector2[1] * rhs_vector2[1])
			else
				self.raise(error.TypeMismatch, "vector2 types are not multiplicable with {typename}", .{
					.typename = addable.get_typename(),
				});
	}

	pub fn mul_vector3(self: *RuntimeEnv, rhs_vector3: Vector3) RuntimeError!*RuntimeEnv {
		const addable = try self.expect_object(self.pop());

		defer self.release(addable);

		return if (addable.is_vector3()) |lhs_vector3|
				self.new_vector3(lhs_vector3[0] * rhs_vector3[0], lhs_vector3[1] * rhs_vector3[1], lhs_vector3[2] * rhs_vector3[2])
			else
				self.raise(error.TypeMismatch, "vector3 types are not multiplicable with {typename}", .{
					.typename = addable.get_typename(),
				});
	}

	pub fn neg(self: *RuntimeEnv) RuntimeError!*RuntimeEnv {
		const negatable = try self.expect_object(self.pop());

		defer self.release(negatable);

		return switch (negatable.internal().payload) {
			.fixed => |fixed| self.new_fixed(-fixed),
			.float => |float| self.new_float(-float),

			else => self.raise(error.TypeMismatch, "{typename} is not scalar negatable", .{
				.typename = negatable.get_typename(),
			}),
		};
	}

	pub fn new_boolean(self: *RuntimeEnv, value: bool) RuntimeError!*RuntimeEnv {
		const boolean = try RuntimeObj.allocate(self.allocator, .{
			.ref_count = 1,
			.payload = if (value) .true else .false,
		});

		errdefer self.release(boolean);

		try self.locals.push_one(boolean);

		return self;
	}

	pub fn new_boxed(self: *RuntimeEnv, value: ?*RuntimeObj) RuntimeError!*RuntimeEnv {
		const boxed = try RuntimeObj.allocate(self.allocator, .{
			.ref_count = 1,
			.payload = .{.boxed = if (value) |ref| ref.acquire() else null},
		});

		errdefer self.release(boxed);

		try self.locals.push_one(boxed);

		return self;
	}

	pub fn new_dynamic(self: *RuntimeEnv, userdata: []const coral.io.Byte, typeinfo: *const Typeinfo) RuntimeError!*RuntimeEnv {
		coral.debug.assert(userdata.len == typeinfo.size);

		const dynamic = new: {
			const dynamic = try self.allocator.reallocate(null, @sizeOf(usize) + typeinfo.size);

			errdefer self.allocator.deallocate(dynamic);

			coral.io.copy(dynamic, coral.io.bytes_of(&typeinfo));
			coral.io.copy(dynamic[@sizeOf(usize) ..], userdata[0 .. typeinfo.size]);

			break: new try RuntimeObj.allocate(self.allocator, .{
				.ref_count = 1,
				.payload = .{.dynamic = .{.ptr = dynamic.ptr}},
			});
		};

		errdefer self.release(dynamic);

		try self.locals.push_one(dynamic);

		return self;
	}

	pub fn new_fixed(self: *RuntimeEnv, value: Fixed) RuntimeError!*RuntimeEnv {
		const syscall = try RuntimeObj.allocate(self.allocator, .{
			.ref_count = 1,
			.payload = .{.fixed = value},
		});

		errdefer self.release(syscall);

		try self.locals.push_one(syscall);

		return self;
	}

	pub fn new_float(self: *RuntimeEnv, value: Float) RuntimeError!*RuntimeEnv {
		const syscall = try RuntimeObj.allocate(self.allocator, .{
			.ref_count = 1,
			.payload = .{.float = value},
		});

		errdefer self.release(syscall);

		try self.locals.push_one(syscall);

		return self;
	}

	pub fn new_string(self: *RuntimeEnv, value: []const coral.io.Byte) RuntimeError!*RuntimeEnv {
		if (value.len > coral.math.max_int(@typeInfo(Fixed).Int)) {
			return error.OutOfMemory;
		}

		const string = new: {
			const string = try coral.io.allocate_copy(coral.io.Byte, self.allocator, value);

			errdefer self.allocator.deallocate(string);

			break: new try RuntimeObj.allocate(self.allocator, .{
				.payload = .{
					.string = .{
						.ptr = string.ptr,
						.len = @intCast(string.len),
					},
				},

				.ref_count = 1,
			});
		};

		errdefer self.release(string);

		try self.locals.push_one(string);

		return self;
	}

	pub fn new_syscall(self: *RuntimeEnv, value: *const Syscall) RuntimeError!*RuntimeEnv {
		const syscall = try RuntimeObj.allocate(self.allocator, .{
			.ref_count = 1,
			.payload = .{.syscall = value},
		});

		errdefer self.release(syscall);

		try self.locals.push_one(syscall);

		return self;
	}

	pub fn new_table(self: *RuntimeEnv) RuntimeError!*RuntimeEnv {
		var table = Table.init(self);

		errdefer table.deinit(self);

		return self.new_dynamic(coral.io.bytes_of(&table), Table.typeinfo);
	}

	pub fn new_vector2(self: *RuntimeEnv, x: f32, y: f32) RuntimeError!*RuntimeEnv {
		const vector2 = try RuntimeObj.allocate(self.allocator, .{
			.ref_count = 1,
			.payload = .{.vector2 = .{x, y}},
		});

		errdefer self.release(vector2);

		try self.locals.push_one(vector2);

		return self;
	}

	pub fn new_vector3(self: *RuntimeEnv, x: f32, y: f32, z: f32) RuntimeError!*RuntimeEnv {
		const vector3 = try RuntimeObj.allocate(self.allocator, .{
			.ref_count = 1,
			.payload = .{.vector3 = .{x, y, z}},
		});

		errdefer self.release(vector3);

		try self.locals.push_one(vector3);

		return self;
	}

	pub fn new_symbol(self: *RuntimeEnv, value: []const coral.io.Byte) RuntimeError!*RuntimeEnv {
		const symbol = try RuntimeObj.allocate(self.allocator, .{
			.payload = .{
				.symbol = self.interned_symbols.lookup(value) orelse create: {
					const symbol_string = try coral.io.allocate_string(self.allocator, value);

					errdefer self.allocator.deallocate(symbol_string);

					coral.debug.assert(try self.interned_symbols.insert(symbol_string[0 .. value.len], symbol_string));

					break: create symbol_string;
				},
			},

			.ref_count = 1,
		});

		errdefer self.release(symbol);

		try self.locals.push_one(symbol);

		return self;
	}

	pub fn pop(self: *RuntimeEnv) ?*RuntimeObj {
		const popped = self.locals.pop();

		coral.debug.assert(popped != null);

		return popped.?;
	}

	pub fn print(self: *RuntimeEnv, buffer: []const coral.io.Byte) void {
		if (self.options.print) |op| {
			op(buffer);
		}
	}

	pub fn print_error(self: *RuntimeEnv, buffer: []const coral.io.Byte) void {
		if (self.options.print_error) |op| {
			op(buffer);
		}
	}

	pub fn push(self: *RuntimeEnv, value: ?*RuntimeObj) RuntimeError!*RuntimeEnv {
		if (value) |object| {
			const acquired = object.acquire();

			errdefer self.release(acquired);

			try self.locals.push_one(acquired);
		} else {
			try self.locals.push_one(null);
		}

		return self;
	}

	pub fn raise(self: *RuntimeEnv, error_value: RuntimeError, comptime format: []const coral.io.Byte, args: anytype) RuntimeError {
		const formatted_message = try coral.utf8.alloc_formatted(self.allocator, format, args);

		defer self.allocator.deallocate(formatted_message);

		self.print_error(formatted_message);

		if (!self.frames.is_empty()) {
			self.print_error("stack trace:");

			var remaining_frames = self.frames.values.len;

			while (remaining_frames != 0) {
				remaining_frames -= 1;

				const callable = self.frames.values[remaining_frames].callable;
				const name = (try (try self.push(callable)).to_string()).pop().?;

				defer self.release(name);

				if (callable.is_dynamic(Chunk.typeinfo)) |chunk_userdata| {
					const chunk = @as(*Chunk, @ptrCast(@alignCast(chunk_userdata)));
					const line = chunk.lines.values[chunk.cursor];

					const chunk_name = try coral.utf8.alloc_formatted(self.allocator, "{name}@{line_number}", .{
						.name = get_name: {
							const string = name.is_string();

							coral.debug.assert(string != null);

							break: get_name string.?;
						},

						.line_number = line.number,
					});

					defer self.allocator.deallocate(chunk_name);

					self.print_error(chunk_name);
				} else {
					self.print_error(get_name: {
						const string = name.is_string();

						coral.debug.assert(string != null);

						break: get_name string.?;
					});
				}
			}
		}

		return error_value;
	}

	pub fn release(self: *RuntimeEnv, value: *RuntimeObj) void {
		coral.debug.assert(value.internal().ref_count != 0);

		value.internal().ref_count -= 1;

		if (value.internal().ref_count == 0) {
			switch (value.internal().payload) {
				.false, .true, .float, .fixed, .symbol, .vector2, .vector3, .syscall => {},

				.boxed => |*boxed| {
					if (boxed.*) |boxed_value| {
						self.release(boxed_value);
					}
				},

				.string => |string| {
					coral.debug.assert(string.len >= 0);
					self.allocator.deallocate(string.ptr[0 .. @intCast(string.len)]);
				},

				.dynamic => |dynamic| {
					if (dynamic.typeinfo().destruct) |destruct| {
						destruct(self, dynamic.userdata());
					}

					self.allocator.deallocate(dynamic.unpack());
				},
			}

			self.allocator.deallocate(value.internal());
		}
	}

	pub fn sub_fixed(self: *RuntimeEnv, rhs_fixed: Fixed) RuntimeError!*RuntimeEnv {
		const addable = try self.expect_object(self.pop());

		defer self.release(addable);

		return switch (addable.internal().payload) {
			.fixed => |lhs_fixed| if (coral.math.checked_sub(lhs_fixed, rhs_fixed)) |result|
					self.new_fixed(result)
				else
					self.new_float(@as(Float, @floatFromInt(lhs_fixed)) - @as(Float, @floatFromInt(rhs_fixed))),

			.float => |lhs_float| self.new_float(lhs_float + @as(Float, @floatFromInt(rhs_fixed))),

			.vector2 => |lhs_vector2| new: {
				const rhs_scalar = @as(f32, @floatFromInt(rhs_fixed));

				break: new self.new_vector2(lhs_vector2[0] - rhs_scalar, lhs_vector2[1] - rhs_scalar);
			},

			.vector3 => |lhs_vector3| new: {
				const rhs_scalar = @as(f32, @floatFromInt(rhs_fixed));

				break: new self.new_vector3(lhs_vector3[0] - rhs_scalar, lhs_vector3[1] - rhs_scalar, lhs_vector3[2] - rhs_scalar);
			},

			else => self.raise(error.TypeMismatch, "fixed types are not multiplicable with {typename}", .{
				.typename = addable.get_typename(),
			}),
		};
	}

	pub fn sub_float(self: *RuntimeEnv, rhs_float: Float) RuntimeError!*RuntimeEnv {
		const addable = try self.expect_object(self.pop());

		defer self.release(addable);

		return switch (addable.internal().payload) {
			.fixed => |lhs_fixed| self.new_float(@as(Float, @floatFromInt(lhs_fixed)) - rhs_float),
			.float => |lhs_float| self.new_float(lhs_float - lhs_float),

			.vector2 => |lhs_vector2| new: {
				const rhs_scalar = @as(f32, @floatCast(rhs_float));

				break: new self.new_vector2(lhs_vector2[0] - rhs_scalar, lhs_vector2[1] - rhs_scalar);
			},

			.vector3 => |lhs_vector3| new: {
				const rhs_scalar = @as(f32, @floatCast(rhs_float));

				break: new self.new_vector3(lhs_vector3[0] - rhs_scalar, lhs_vector3[1] - rhs_scalar, lhs_vector3[2] - rhs_scalar);
			},

			else => self.raise(error.TypeMismatch, "fixed types are not multiplicable with {typename}", .{
				.typename = addable.get_typename(),
			}),
		};
	}

	pub fn sub_vector2(self: *RuntimeEnv, rhs_vector2: Vector2) RuntimeError!*RuntimeEnv {
		const addable = try self.expect_object(self.pop());

		defer self.release(addable);

		return if (addable.is_vector2()) |lhs_vector2|
				self.new_vector2(lhs_vector2[0] - rhs_vector2[0], lhs_vector2[1] - rhs_vector2[1])
			else
				self.raise(error.TypeMismatch, "vector2 types are not multiplicable with {typename}", .{
					.typename = addable.get_typename(),
				});
	}

	pub fn sub_vector3(self: *RuntimeEnv, rhs_vector3: Vector3) RuntimeError!*RuntimeEnv {
		const addable = try self.expect_object(self.pop());

		defer self.release(addable);

		return if (addable.is_vector3()) |lhs_vector3|
				self.new_vector3(lhs_vector3[0] - rhs_vector3[0], lhs_vector3[1] - rhs_vector3[1], lhs_vector3[2] - rhs_vector3[2])
			else
				self.raise(error.TypeMismatch, "vector3 types are not multiplicable with {typename}", .{
					.typename = addable.get_typename(),
				});
	}

	pub fn to_string(self: *RuntimeEnv) RuntimeError!*RuntimeEnv {
		const decimal_format = coral.utf8.DecimalFormat.default;
		const value = try self.expect_object(self.pop());

		defer self.release(value);

		return switch (value.internal().payload) {
			.false => self.new_string("false"),
			.true => self.new_string("true"),

			.fixed => |fixed| convert: {
				var string = [_:0]coral.io.Byte{0} ** 32;
				var buffer = coral.io.FixedBuffer{.bytes = &string};
				const length = decimal_format.print(buffer.as_writer(), fixed);

				coral.debug.assert(length != null);

				break: convert self.new_string(string[0 .. length.?]);
			},

			.float => |float| convert: {
				var string = [_:0]coral.io.Byte{0} ** 32;
				var buffer = coral.io.FixedBuffer{.bytes = &string};
				const length = decimal_format.print(buffer.as_writer(), float);

				coral.debug.assert(length != null);

				break: convert self.new_string(string[0 .. length.?]);
			},

			.boxed => |boxed| (try self.push(try self.expect_object(boxed))).to_string(),
			.symbol => |symbol| self.new_string(coral.io.slice_sentineled(@as(coral.io.Byte, 0), symbol)),

			.string => acquire: {
				try self.locals.push_one(value.acquire());

				break: acquire self;
			},

			.vector2 => |vector2| convert: {
				var string = [_:0]coral.io.Byte{0} ** 64;
				var buffer = coral.io.FixedBuffer{.bytes = &string};

				const length = coral.utf8.print_formatted(buffer.as_writer(), "@vec2({x}, {y})", .{
					.x = vector2[0],
					.y = vector2[1],
				});

				coral.debug.assert(length != null);

				break: convert self.new_string(string[0 .. length.?]);
			},

			.vector3 => |vector3| convert: {
				var string = [_:0]coral.io.Byte{0} ** 96;
				var buffer = coral.io.FixedBuffer{.bytes = &string};

				const length = coral.utf8.print_formatted(buffer.as_writer(), "@vec3({x}, {y}, {z})", .{
					.x = vector3[0],
					.y = vector3[1],
					.z = vector3[2],
				});

				coral.debug.assert(length != null);

				break: convert self.new_string(string[0 .. length.?]);
			},

			.syscall => self.new_string("syscall"),

			.dynamic => |dynamic| dynamic_to_string: {
				const string = try dynamic.typeinfo().to_string(self, dynamic.userdata());

				errdefer self.release(string);

				try self.locals.push_one(string);

				break: dynamic_to_string self;
			},
		};
	}

	pub fn boxed_get(self: *RuntimeEnv) RuntimeError!*RuntimeEnv {
		const unboxable = try self.expect_object(self.pop());

		defer self.release(unboxable);

		return switch (unboxable.internal().payload) {
			.boxed => |boxed| self.push(boxed),

			else => self.raise(error.TypeMismatch, "{typename} is not unboxable", .{
				.typename = unboxable.get_typename(),
			}),
		};
	}

	pub fn boxed_set(self: *RuntimeEnv, value: ?*RuntimeObj) RuntimeError!void {
		const unboxable = try self.expect_object(self.pop());

		defer self.release(unboxable);

		switch (unboxable.internal().payload) {
			.boxed => |*unboxed_value| {
				if (unboxed_value.*) |unboxed_object| {
					self.release(unboxed_object);
				}

				unboxed_value.* = if (value) |object| object.acquire() else null;
			},

			else => return self.raise(error.TypeMismatch, "{typename} is not unboxable", .{
				.typename = unboxable.get_typename(),
			}),
		}
	}
};

pub const RuntimeError = coral.io.AllocationError || error {
	IllegalState,
	TypeMismatch,
	BadOperation,
	BadSyntax,
};

pub const RuntimeObj = opaque {
	const Internal = struct {
		ref_count: u16,

		payload: union (enum) {
			false,
			true,
			float: Float,
			fixed: Fixed,
			symbol: [*:0]const coral.io.Byte,
			vector2: Vector2,
			vector3: Vector3,
			syscall: *const Syscall,
			boxed: ?*RuntimeObj,

			string: struct {
				ptr: [*]coral.io.Byte,
				len: Fixed,

				const Self = @This();

				fn unpack(self: Self) []coral.io.Byte {
					coral.debug.assert(self.len >= 0);

					return self.ptr[0 .. @intCast(self.len)];
				}
			},

			dynamic: struct {
				ptr: [*]coral.io.Byte,

				const Self = @This();

				fn typeinfo(self: Self) *const Typeinfo {
					return @as(**const Typeinfo, @ptrCast(@alignCast(self.ptr))).*;
				}

				fn unpack(self: Self) []coral.io.Byte {
					return self.ptr[0 .. (@sizeOf(usize) + self.typeinfo().size)];
				}

				fn userdata(self: Self) []coral.io.Byte {
					const unpacked = self.unpack();
					const address_size = @sizeOf(usize);

					coral.debug.assert(unpacked.len >= address_size);

					return unpacked[address_size ..];
				}
			},
		},
	};

	pub fn acquire(self: *const RuntimeObj) *RuntimeObj {
		const self_object = self.internal();

		self_object.ref_count += 1;

		return@ptrCast(self_object);
	}

	fn allocate(allocator: coral.io.Allocator, data: Internal) coral.io.AllocationError!*RuntimeObj {
		return @ptrCast(try coral.io.allocate_one(allocator, data));
	}

	fn internal(self: *const RuntimeObj) *Internal {
		return @constCast(@ptrCast(@alignCast(self)));
	}

	pub fn equals(self: *RuntimeObj, other: *RuntimeObj) bool {
		return switch (self.internal().payload) {
			.false => other.internal().payload == .false,
			.true => other.internal().payload == .true,

			.fixed => |lhs_fixed| switch (other.internal().payload) {
				.fixed => |rhs_fixed| rhs_fixed == lhs_fixed,
				.float => |rhs_float| rhs_float == @as(Float, @floatFromInt(lhs_fixed)),
				else => false,
			},

			.float => |lhs_float| switch (other.internal().payload) {
				.float => |rhs_float| rhs_float == lhs_float,
				.fixed => |rhs_fixed| @as(Float, @floatFromInt(rhs_fixed)) == lhs_float,
				else => false,
			},

			.symbol => |lhs_symbol| switch (other.internal().payload) {
				.symbol => |rhs_symbol| lhs_symbol == rhs_symbol,
				else => false,
			},

			.boxed => |boxed| if (boxed) |object| self.equals(object) else false,

			.vector2 => |lhs_vector| switch (other.internal().payload) {
				.vector2 => |rhs_vector| lhs_vector[0] == rhs_vector[0] and lhs_vector[1] == rhs_vector[1],
				else => false,
			},

			.vector3 => |lhs_vector| switch (other.internal().payload) {
				.vector3 => |rhs_vector| lhs_vector[0] == rhs_vector[0] and lhs_vector[1] == rhs_vector[1] and lhs_vector[2] == rhs_vector[2],
				else => false,
			},

			.syscall => |lhs_syscall| switch (other.internal().payload) {
				.syscall => |rhs_syscall| lhs_syscall == rhs_syscall,
				else => false,
			},

			.string => |lhs_string| switch (other.internal().payload) {
				.string => |rhs_string| coral.io.are_equal(lhs_string.unpack(), rhs_string.unpack()),
				else => false,
			},

			.dynamic => |lhs_dynamic| switch (other.internal().payload) {
				.dynamic => |rhs_dynamic|
					lhs_dynamic.typeinfo() == rhs_dynamic.typeinfo() and
					lhs_dynamic.userdata().ptr == rhs_dynamic.userdata().ptr,

				else => false,
			},
		};
	}

	pub fn get_hash(self: *RuntimeObj) usize {
		return switch (self.internal().payload) {
			.false => 1237,
			.true => 1231,
			.float => |float| @bitCast(float),
			.fixed => |fixed| @intCast(@as(u32, @bitCast(fixed))),
			.symbol => |symbol| @intFromPtr(symbol),
			.vector2 => |vector| @bitCast(vector),
			.vector3 => |vector| coral.io.jenkins_hash(@typeInfo(usize).Int, coral.io.bytes_of(&vector)),
			.syscall => |syscall| @intFromPtr(syscall),
			.boxed => |boxed| @intFromPtr(boxed),
			.string => |string| coral.io.djb2_hash(@typeInfo(usize).Int, string.unpack()),
			.dynamic => |dynamic| @intFromPtr(dynamic.typeinfo()) ^ @intFromPtr(dynamic.userdata().ptr),
		};
	}

	pub fn get_typename(self: *RuntimeObj) []const coral.io.Byte {
		return switch (self.internal().payload) {
			.false => "false",
			.true => "true",
			.float => "float",
			.fixed => "fixed",
			.symbol => "symbol",
			.vector2 => "vector2",
			.vector3 => "vector3",
			.syscall => "syscall",
			.boxed => "boxed",
			.string => "string",
			.dynamic => "dynamic",
		};
	}

	pub fn is_dynamic(self: *RuntimeObj, typeinfo: *const Typeinfo) ?[]u8 {
		return switch (self.internal().payload) {
			.dynamic => |dynamic| if (dynamic.typeinfo() == typeinfo) dynamic.userdata() else null,
			else => null,
		};
	}

	pub fn is_false(self: *RuntimeObj) bool {
		return !self.is_true();
	}

	pub fn is_fixed(self: *RuntimeObj) ?Fixed {
		return switch (self.internal().payload) {
			.fixed => |fixed| fixed,
			else => null,
		};
	}

	pub fn is_numeric(self: *RuntimeObj) ?Numeric {
		return switch (self.internal().payload) {
			.fixed => |fixed| .{.fixed = fixed},
			.float => |float| .{.float = float},
			.vector2 => |vector2| .{.vector2 = vector2},
			.vector3 => |vector3| .{.vector3 = vector3},
			else => null,
		};
	}

	pub fn is_string(self: *RuntimeObj) ?[]const coral.io.Byte {
		return switch (self.internal().payload) {
			.string => |string| get: {
				coral.debug.assert(string.len > -1);

				break: get string.ptr[0 .. @intCast(string.len)];
			},

			.symbol => |symbol| coral.io.slice_sentineled(@as(u8, 0), symbol),
			else => null,
		};
	}

	pub fn is_symbol(self: *RuntimeObj) ?[*:0]const coral.io.Byte {
		return switch (self.internal().payload) {
			.symbol => |symbol| symbol,
			else => null,
		};
	}

	pub fn is_true(self: *RuntimeObj) bool {
		return switch (self.internal().payload) {
			.false => false,
			.true => true,
			.float => |float| float != 0,
			.fixed => |fixed| fixed != 0,
			.symbol => true,
			.vector2 => |vector| coral.io.all_equals(coral.io.bytes_of(&vector), 0),
			.vector3 => |vector| coral.io.all_equals(coral.io.bytes_of(&vector), 0),
			.syscall => true,
			.boxed => |boxed| boxed != null,
			.string => |string| string.len != 0,
			.dynamic => true,
		};
	}

	pub fn is_vector2(self: *RuntimeObj) ?Vector2 {
		return switch (self.internal().payload) {
			.vector2 => |vector2| vector2,
			else => null,
		};
	}

	pub fn is_vector3(self: *RuntimeObj) ?Vector3 {
		return switch (self.internal().payload) {
			.vector3 => |vector3| vector3,
			else => null,
		};
	}
};

pub const Syscall = fn (env: *RuntimeEnv) RuntimeError!?*RuntimeObj;

pub const Typeinfo = struct {
	name: []const coral.io.Byte,
	size: usize,
	destruct: ?*const fn (env: *RuntimeEnv, userdata: []coral.io.Byte) void = null,
	to_string: *const fn (env: *RuntimeEnv, userdata: []coral.io.Byte) RuntimeError!*RuntimeObj = default_to_string,
	call: *const fn (env: *RuntimeEnv, userdata: []coral.io.Byte) RuntimeError!?*RuntimeObj = default_call,
	get: *const fn (env: *RuntimeEnv, userdata: []coral.io.Byte, index: *RuntimeObj) RuntimeError!?*RuntimeObj = default_get,
	set: *const fn (env: *RuntimeEnv, userdata: []coral.io.Byte, value: *RuntimeObj, value: ?*RuntimeObj) RuntimeError!void = default_set,

	fn default_call(env: *RuntimeEnv, _: []coral.io.Byte) RuntimeError!?*RuntimeObj {
		return env.raise(error.BadOperation, "this dynamic object is not callable", .{});
	}

	fn default_get(env: *RuntimeEnv, _: []coral.io.Byte, _: *RuntimeObj) RuntimeError!?*RuntimeObj {
		return env.raise(error.BadOperation, "this dynamic object is not get-indexable", .{});
	}

	fn default_set(env: *RuntimeEnv, _: []coral.io.Byte, _: *RuntimeObj, _: ?*RuntimeObj) RuntimeError!void {
		return env.raise(error.BadOperation, "this dynamic object is not set-indexable", .{});
	}

	fn default_to_string(env: *RuntimeEnv, _: []coral.io.Byte) RuntimeError!*RuntimeObj {
		return (try env.new_string("{}")).pop().?;
	}
};

pub const Vector2 = [2]f32;

pub const Vector3 = [3]f32;

pub fn get_at(env: *RuntimeEnv, indexable: *RuntimeObj, index: Fixed) RuntimeError!?*RuntimeObj {
	const index_number = (try env.new_fixed(index)).pop();

	defer env.discard(index_number);

	return (try (try env.push(indexable)).index_get(index_number)).pop();
}

pub fn get_field(env: *RuntimeEnv, indexable: *RuntimeObj, field: []const coral.io.Byte) RuntimeError!?*RuntimeObj {
	const field_symbol = (try env.new_symbol(field)).pop().?;

	defer env.release(field_symbol);

	return (try (try env.push(indexable)).index_get(field_symbol)).pop();
}

pub fn get_key(env: *RuntimeEnv, indexable: *RuntimeObj, key: []const coral.io.Byte) RuntimeError!?*RuntimeObj {
	const key_string = (try env.new_string(key)).pop();

	defer env.discard(key_string);

	return (try (try env.push(indexable)).index_get(key_string)).pop();
}
